/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_OTW_DEBUG	FALSE

#include "grass/grass.pmh"

using namespace fe;

void OperatorTwistWrap::initialize(void)
{
	catalog<String>("Attach Attribute")="twistAttach";
	catalog<String>("Facing Attribute")="facing";
	catalog<Real>("Facing Twist")=1.0;
	catalog<Real>("Facing Twist","max")=10.0;
	catalog< sp<Component> >("Reference Mesh");
	catalog< sp<Component> >("Reference Curves");
	catalog< sp<Component> >("Deformed Curves");

	catalog< sp<Component> >("Output Mesh");
	catalog<String>("Output Mesh","IO")="output";
	catalog<String>("Output Mesh","copy")="Reference Mesh";
	catalog<bool>("Output Mesh","recycle")=true;
}

void OperatorTwistWrap::handle(Record& a_rSignal)
{
#if FE_OTW_DEBUG
	feLog("OperatorTwistWrap::handle\n");
#endif

	m_facingName=catalog<String>("Facing Attribute");
	m_attachName=catalog<String>("Attach Attribute");

	//* access output curves (copy of input surface)
	m_spOutput=surfaceOutput(a_rSignal);
	if(m_spOutput.isNull())
	{
		feLog("OperatorTwistWrap::handle"
				" output surface not SurfaceAccessibleI\n");
		return;
	}

	//* access reference curves
	m_spInput=catalog< sp<Component> >("Reference Mesh");
	if(m_spInput.isNull())
	{
		feLog("OperatorTwistWrap::handle"
				" reference mesh not SurfaceAccessibleI\n");
		return;
	}

	//* access reference curves
	m_spReference=catalog< sp<Component> >("Reference Curves");
	if(m_spReference.isNull())
	{
		feLog("OperatorTwistWrap::handle"
				" reference curves not SurfaceAccessibleI\n");
		return;
	}

	//* access deformed curves
	m_spDeformed=catalog< sp<Component> >("Deformed Curves");
	if(m_spDeformed.isNull())
	{
		feLog("OperatorTwistWrap::handle"
				" deformed curves not SurfaceAccessibleI\n");
		return;
	}

	sp<SurfaceAccessorI> spAccessorOutput=m_spOutput->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	FEASSERT(spAccessorOutput.isValid());

	sp<SurfaceAccessorI> spAccessorRef=m_spReference->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	if(spAccessorRef.isNull())
	{
		feLog("OperatorTwistWrap::handle"
				" reference curves failed primitive vertex access\n");
		return;
	}

	sp<SurfaceAccessorI> spAccessorDef=m_spDeformed->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	FEASSERT(spAccessorDef.isValid());
	if(spAccessorDef.isNull())
	{
		feLog("OperatorTwistWrap::handle"
				" deformed curves failed primitive vertex access\n");
		return;
	}

	const U32 primitiveCount = spAccessorOutput->count();
	const U32 primitiveCountRef = spAccessorRef->count();
	const U32 primitiveCountDef = spAccessorRef->count();
	U32 primitiveCountMin = primitiveCountRef<primitiveCountDef?
			primitiveCountRef: primitiveCountDef;
	primitiveCountMin = primitiveCountMin<primitiveCount?
			primitiveCountMin: primitiveCount;

	setFullRange(0,primitiveCountMin-1);
	OperatorThreaded::handle(a_rSignal);

#if FE_OTW_DEBUG
	feLog("OperatorTwistWrap::handle done\n");
#endif
}

void OperatorTwistWrap::run(I32 a_id,Range a_range)
{
	const I32 start=a_range.m_start;
	const I32 end=a_range.m_end;

	Real facingTwist=catalog<Real>("Facing Twist");

#if FE_OTW_DEBUG
	feLog("OperatorTwistWrap::run running thread %d [%d,%d]\n",a_id,start,end);
#endif

	sp<SurfaceAccessorI> spAccessorInput=m_spInput->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	FEASSERT(spAccessorInput.isValid());

	sp<SurfaceAccessorI> spAccessorOutput=m_spOutput->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	FEASSERT(spAccessorOutput.isValid());

	sp<SurfaceAccessorI> spAccessorAttach=m_spOutput->accessor(
			SurfaceAccessibleI::e_point,m_attachName);
	FEASSERT(spAccessorAttach.isValid());

	sp<SurfaceAccessorI> spAccessorRef=m_spReference->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	if(spAccessorRef.isNull())
	{
		feLog("OperatorTwistWrap::handle"
				" reference curves failed primitive vertex access\n");
		return;
	}

	sp<SurfaceAccessorI> spAccessorDef=m_spDeformed->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	FEASSERT(spAccessorDef.isValid());
	if(spAccessorDef.isNull())
	{
		feLog("OperatorTwistWrap::handle"
				" deformed curves failed primitive vertex access\n");
		return;
	}

	sp<SurfaceAccessorI> spAccessorFacingRef=m_spReference->accessor(
			SurfaceAccessibleI::e_point,m_facingName);
	FEASSERT(spAccessorFacingRef.isValid());
	if(spAccessorRef.isNull())
	{
		feLog("OperatorTwistWrap::handle"
				" reference curves failed point 'facing' access\n");
		return;
	}

	sp<SurfaceAccessorI> spAccessorFacingDef=m_spDeformed->accessor(
			SurfaceAccessibleI::e_point,m_facingName);
	FEASSERT(spAccessorFacingDef.isValid());
	if(spAccessorRef.isNull())
	{
		feLog("OperatorTwistWrap::handle"
				" deformed curves failed point 'facing' access\n");
		return;
	}

	for(U32 index=start;index<=end;index++)
	{
		const U32 subCountRef = spAccessorRef->subCount(index);
		const U32 subCountDef = spAccessorDef->subCount(index);
		if(!subCountRef || subCountRef!=subCountDef)
		{
			feLog("OperatorTwistWrap::handle"
					" primitive %d mismatched subCount: %d ref %d def\n",
					index,subCountRef,subCountDef);
			continue;
		}

		SpatialTransform transformArray[subCountRef];

		SpatialVector pointRef(0.0,0.0,0.0);
		SpatialVector pointDef(0.0,0.0,0.0);
		SpatialVector pointRefPost=spAccessorRef->spatialVector(index,0);
		SpatialVector pointDefPost=spAccessorDef->spatialVector(index,0);
		for(U32 subIndex=0;subIndex<subCountRef;subIndex++)
		{
			const U32 vertexIndexRef=spAccessorRef->integer(index,subIndex);
			const U32 vertexIndexDef=spAccessorDef->integer(index,subIndex);

			const SpatialVector pointRefPre=pointRef;
			const SpatialVector pointDefPre=pointDef;
			pointRef=pointRefPost;
			pointDef=pointDefPost;

			pointRefPost=spAccessorRef->spatialVector(index,subIndex+1);
			pointDefPost=spAccessorDef->spatialVector(index,subIndex+1);

			const SpatialVector facingRef=
					spAccessorFacingRef->spatialVector(vertexIndexRef);
			SpatialVector facingDef=
					spAccessorFacingDef->spatialVector(vertexIndexDef);

			SpatialVector dirRef;
			SpatialVector dirDef;
			if(subIndex==0)
			{
				dirRef=unit(pointRefPost-pointRef);
				dirDef=unit(pointDefPost-pointDef);
			}
			else if(subIndex==subCountRef-1)
			{
				dirRef=unit(pointRef-pointRefPre);
				dirDef=unit(pointDef-pointDefPre);
			}
			else
			{
				dirRef=unit(pointRefPost-pointRefPre);
				dirDef=unit(pointDefPost-pointDefPre);
			}

			//* HACK move facing in direction of change
			SpatialVector change=pointDef-pointRef;
			facingDef+=facingTwist*change*SpatialVector(1.0,0.0,1.0);
			facingDef[1]+=facingTwist*(fabs(change[0])+fabs(change[2]));
			normalize(facingDef);

			SpatialTransform frameRef;
			BWORD success=makeFrame(frameRef,pointRef,dirRef,facingRef);

			SpatialTransform frameDef;
			success=makeFrame(frameDef,pointDef,dirDef,facingDef);

			SpatialTransform frameRefInv;
			invert(frameRefInv,frameRef);

			transformArray[subIndex]=frameRefInv*frameDef;
		}

		const U32 subCount=spAccessorOutput->subCount(index);
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const U32 vertexIndex=spAccessorOutput->integer(index,subIndex);
			const Real attach=spAccessorAttach->real(vertexIndex);

			const SpatialVector point=
					spAccessorInput->spatialVector(index,subIndex);

			U32 first=U32(attach);
			Real between=attach-Real(first);

			const SpatialVector point0=
					transformVector(transformArray[first],point);
			if(between<1e-6)
			{
				spAccessorOutput->set(index,subIndex,point0);
				continue;
			}

			const SpatialVector point1=
					transformVector(transformArray[first+1],point);
			const SpatialVector blend=(1.0-between)*point0+between*point1;

			spAccessorOutput->set(index,subIndex,blend);
		}
	}

#if FE_OTW_DEBUG
	feLog("OperatorTwistWrap::run done running thread %d\n",a_id);
#endif
}
